/**************************************************************************/
/* File: ioctl.c - CAN driver IOCTL functions support                     */
/*                                                                        */
/* LinCAN - (Not only) Linux CAN bus driver                               */
/* Copyright (C) 2002-2009 DCE FEE CTU Prague <http://dce.felk.cvut.cz>   */
/* Copyright (C) 2002-2009 Pavel Pisa <pisa@cmp.felk.cvut.cz>             */
/* Funded by OCERA and FRESCOR IST projects                               */
/* Based on CAN driver code by Arnaud Westenberg <arnaud@wanadoo.nl>      */
/*                                                                        */
/* LinCAN is free software; you can redistribute it and/or modify it      */
/* under terms of the GNU General Public License as published by the      */
/* Free Software Foundation; either version 2, or (at your option) any    */
/* later version.  LinCAN is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty    */
/* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    */
/* General Public License for more details. You should have received a    */
/* copy of the GNU General Public License along with LinCAN; see file     */
/* COPYING. If not, write to the Free Software Foundation, 675 Mass Ave,  */
/* Cambridge, MA 02139, USA.                                              */
/*                                                                        */
/* To allow use of LinCAN in the compact embedded systems firmware        */
/* and RT-executives (RTEMS for example), main authors agree with next    */
/* special exception:                                                     */
/*                                                                        */
/* Including LinCAN header files in a file, instantiating LinCAN generics */
/* or templates, or linking other files with LinCAN objects to produce    */
/* an application image/executable, does not by itself cause the          */
/* resulting application image/executable to be covered by                */
/* the GNU General Public License.                                        */
/* This exception does not however invalidate any other reasons           */
/* why the executable file might be covered by the GNU Public License.    */
/* Publication of enhanced or derived LinCAN files is required although.  */
/**************************************************************************/

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/ioctl.h"

int can_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int i=0;
	unsigned short channel=0;
	unsigned btr0=0, btr1=0;
	struct canuser_t *canuser = (struct canuser_t*)(file->private_data);
	struct msgobj_t *obj;
	struct canchip_t *chip;
	struct canque_ends_t *qends;
	
	if(!canuser || (canuser->magic != CAN_USER_MAGIC)){
		CANMSG("can_ioctl: bad canuser magic\n");
		return -ENODEV;
	}
	
	obj = canuser->msgobj;
	if (obj == NULL) {
		CANMSG("Could not assign buffer structure\n");
		return -1;
	}

	qends = canuser->qends;
	chip = obj->hostchip;
	if (chip == NULL) {
		CANMSG("Device is not correctly configured.\n");
		CANMSG("Please reload the driver.\n");
		return -1;
	}

	switch (cmd) {
		case CAN_DRV_QUERY: {
			return can_ioctl_query(canuser, arg);
		}
		case STAT: {
			for (i=0x0; i<0x100; i++)
				CANMSG("0x%x is 0x%x\n",i,can_read_reg(chip,i));
			break;
		}
		case CMD_START: {
			if (chip->chipspecops->start_chip(chip))
				return -1;
			break;
		}
		case CMD_STOP: {
			if (chip->chipspecops->stop_chip(chip))
				return -1;
			break;
		}
		case CANQUE_FLUSH: {
			canque_flush(canuser->rx_edge0);
			break;
		}
		case CONF_FILTER: {

			/* In- and output buffer re-initialization */
			
			if(canuser->rx_edge0){
				canque_set_filt(canuser->rx_edge0, arg, ~0, 0);
			}

			break;
		}
		
		case CANQUE_FILTER: {
			struct canfilt_t canfilt;
			int ret;
			ret = copy_from_user(&canfilt, (void*)arg, sizeof(struct canfilt_t));
			if(ret)	return -EFAULT;
			if(canuser->rx_edge0){
				canque_set_filt(canuser->rx_edge0, canfilt.id, canfilt.mask, canfilt.flags);
			}
			break;
		}
		
		case CANRTR_READ: {
			int ret;
			struct canmsg_t rtr_msg;
			
			ret = copy_from_user(&rtr_msg, (void*)arg, sizeof(struct canmsg_t));
			if(ret)	return -EFAULT;
			ret = can_ioctl_remote_read(canuser, &rtr_msg, rtr_msg.id, 0);
			if(ret<0) return ret;
			ret = copy_to_user((void*)arg, &rtr_msg, sizeof(struct canmsg_t));
			if(ret)	return -EFAULT;
			break;
		}

		case CONF_BAUD: {
			channel = arg & 0xff;
			btr0 = (arg >> 8) & 0xff;
			btr1 = (arg >> 16) & 0xff;

			if (chip->chipspecops->set_btregs(chip, btr0, btr1)) {
				CANMSG("Error setting bit timing registers\n");
				return -1;
			}
			break;
		}
		
		case CONF_BAUDPARAMS: {
			struct can_baudparams_t params;
			int ret;
			
			ret = copy_from_user(&params, (void*)arg, sizeof(struct can_baudparams_t));
			if(ret)	return -EFAULT;

			if(params.flags == -1) params.flags = 0;
			if(params.baudrate == -1) params.baudrate = chip->baudrate;
			if(params.sjw == -1) params.sjw = 0;
			if(params.sample_pt == -1) params.sample_pt = 75;
			i=chip->chipspecops->baud_rate(chip, params.baudrate, chip->clock, params.sjw,
                                                        params.sample_pt, params.flags);
			if(i>=0) chip->baudrate = params.baudrate;
			else {
				CANMSG("Error setting baud parameters\n");
				return -1;
			}
			break;
		}


		default: {
			CANMSG("Not a valid ioctl command\n");
			return -EINVAL;
		}
		
	}

	return 0;
}
