/**************************************************************************/
/* File: write.c - systemcall to send messages through driver FIFOs       */
/*                                                                        */
/* LinCAN - (Not only) Linux CAN bus driver                               */
/* Copyright (C) 2002-2009 DCE FEE CTU Prague <http://dce.felk.cvut.cz>   */
/* Copyright (C) 2002-2009 Pavel Pisa <pisa@cmp.felk.cvut.cz>             */
/* Funded by OCERA and FRESCOR IST projects                               */
/* Based on CAN driver code by Arnaud Westenberg <arnaud@wanadoo.nl>      */
/*                                                                        */
/* LinCAN is free software; you can redistribute it and/or modify it      */
/* under terms of the GNU General Public License as published by the      */
/* Free Software Foundation; either version 2, or (at your option) any    */
/* later version.  LinCAN is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty    */
/* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    */
/* General Public License for more details. You should have received a    */
/* copy of the GNU General Public License along with LinCAN; see file     */
/* COPYING. If not, write to the Free Software Foundation, 675 Mass Ave,  */
/* Cambridge, MA 02139, USA.                                              */
/*                                                                        */
/* To allow use of LinCAN in the compact embedded systems firmware        */
/* and RT-executives (RTEMS for example), main authors agree with next    */
/* special exception:                                                     */
/*                                                                        */
/* Including LinCAN header files in a file, instantiating LinCAN generics */
/* or templates, or linking other files with LinCAN objects to produce    */
/* an application image/executable, does not by itself cause the          */
/* resulting application image/executable to be covered by                */
/* the GNU General Public License.                                        */
/* This exception does not however invalidate any other reasons           */
/* why the executable file might be covered by the GNU Public License.    */
/* Publication of enhanced or derived LinCAN files is required although.  */
/**************************************************************************/

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/write.h"

ssize_t can_write(struct file *file, const char *buffer, size_t length, loff_t *offset)
{
	struct canuser_t *canuser = (struct canuser_t*)(file->private_data);
	struct msgobj_t *obj;
	struct canmsg_t msg_buff;
	struct canque_ends_t *qends;
	struct canque_edge_t *qedge;
	struct canque_slot_t *slot;
	int ret;
	unsigned bytes_to_copy;

	if(!canuser || (canuser->magic != CAN_USER_MAGIC)){
		CANMSG("can_write: bad canuser magic\n");
		return -ENODEV;
	}

	if (length < sizeof(struct canmsg_t)) {
		DEBUGMSG("Trying to write less bytes than a CAN message,\n");
		DEBUGMSG("this will always return 0 !\n");
		return 0;
	}
	if (length > INT_MAX) {
		CANMSG("Trying to write more than is supported.\n");
		return -1;
	}
	if (length % sizeof(struct canmsg_t)) {
		CANMSG("The number of bytes requested to be written is not a multiple of\n");
		CANMSG("'sizeof(struct canmsg_t)', currently this is not allowed.\n");
		return -1;
	} 

	/* Initialize hardware pointers */
	obj = canuser->msgobj;
	if (obj == NULL) {
		CANMSG("Could not assign buffer structure\n");
		return -1;
	}
	qends = canuser->qends;


	/* Prepare first message */
	ret = copy_from_user(&msg_buff, buffer, sizeof(struct canmsg_t));
	if(ret)	return -EFAULT;

	/* Automatic selection of extended format if ID>2047 */
	if (msg_buff.id & ~0x7ffl & MSG_ID_MASK ) msg_buff.flags |= MSG_EXT;
	/* has been dependent on "extended" option */

	/* If the output buffer is full, return immediately in case O_NONBLOCK
	 * has been specified or loop until space becomes available.
	 */
	if ((ret=canque_get_inslot4id(qends, &qedge, &slot, 
			0, msg_buff.id, 0))<0){
		DEBUGMSG("Buffer is full\n");
		if(ret < -1)
			return -EIO;

		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;

		ret=canque_get_inslot4id_wait_kern(qends, &qedge, &slot, 
						0, msg_buff.id, 0);
		if(ret<0) {
			if (signal_pending(current))
				return -EINTR;
			/*if (!can_timeout)*/
			return -EIO;
		}
	}
	slot->msg=msg_buff;
	canque_put_inslot(qends, qedge, slot);
	buffer += sizeof(struct canmsg_t);
	bytes_to_copy = length-sizeof(struct canmsg_t);

	/* 
	 * Try to send more messages
	 */
	while (bytes_to_copy >= sizeof(struct canmsg_t)) {
		bytes_to_copy -= sizeof(struct canmsg_t);
		/* Prepare first message */
		ret = copy_from_user(&msg_buff, buffer, sizeof(struct canmsg_t));
		if(ret)	return -EFAULT;

		/* Automatic selection of extended format if ID>2047 */
		if (msg_buff.id & ~0x7ffl & MSG_ID_MASK ) msg_buff.flags |= MSG_EXT;
		/* has been dependent on "extended" option */

		/* Get slot */
		if(canque_get_inslot4id(qends, &qedge, &slot, 
			0, msg_buff.id, 0) < 0) break;
		slot->msg=msg_buff;
		canque_put_inslot(qends, qedge, slot);
		buffer += sizeof(struct canmsg_t);
	}

        if(file->f_flags & O_SYNC) {
		canque_sync_wait_kern(qends, qedge);
	}
	return length-bytes_to_copy;
}	
